@@ -1,7 +1,7 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2010-2012 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2010-2013 -- leonerd@leonerd.org.uk
 
 package IO::Socket::IP;
 
@@ -9,12 +9,13 @@ use strict;
 use warnings;
 use base qw( IO::Socket );
 
-our $VERSION = '0.08_005';
+our $VERSION = '0.19_001';
 
 use Carp;
 
 use Socket 1.97 qw(
    getaddrinfo getnameinfo
+   sockaddr_family
    AF_INET
    AI_PASSIVE
    IPPROTO_TCP IPPROTO_UDP
@@ -27,7 +28,7 @@ use Socket 1.97 qw(
 my $AF_INET6 = eval { Socket::AF_INET6() }; # may not be defined
 my $AI_ADDRCONFIG = eval { Socket::AI_ADDRCONFIG() } || 0;
 use POSIX qw( dup2 );
-use Errno qw( EINVAL EINPROGRESS );
+use Errno qw( EINVAL EINPROGRESS EISCONN );
 
 use constant HAVE_MSWIN32 => ( $^O eq "MSWin32" );
 
@@ -142,7 +143,7 @@ sub import
    {
       return $can_disable_v6only if defined $can_disable_v6only;
 
-      socket my $testsock, Socket::PF_INET6, SOCK_STREAM, 0 or
+      socket my $testsock, Socket::PF_INET6(), SOCK_STREAM, 0 or
          die "Cannot socket(PF_INET6) - $!";
 
       if( setsockopt $testsock, IPPROTO_IPV6, IPV6_V6ONLY, 0 ) {
@@ -232,6 +233,15 @@ C<'udp'>,C<IPPROTO_UDP>). Normally this will be left undefined, and either
 C<getaddrinfo> or the kernel will choose an appropriate value. May be given
 either in string name or numeric form.
 
+=item GetAddrInfoFlags => INT
+
+More flags to pass to the C<getaddrinfo()> function. If not supplied, a
+default of C<AI_ADDRCONFIG> will be used.
+
+These flags will be combined with C<AI_PASSIVE> if the C<Listen> argument is
+given. For more information see the documentation about C<getaddrinfo()> in
+the L<Socket> module.
+
 =item Listen => INT
 
 If defined, puts the socket into listening mode where new connections can be
@@ -298,7 +308,8 @@ detail.
 
 If neither C<Type> nor C<Proto> hints are provided, a default of
 C<SOCK_STREAM> and C<IPPROTO_TCP> respectively will be set, to maintain
-compatibility with C<IO::Socket::INET>.
+compatibility with C<IO::Socket::INET>. Other named arguments that are not
+recognised are ignored.
 
 If the constructor fails, it will set C<$@> to an appropriate error message;
 this may be from C<$!> or it may be some other string; not every failure
@@ -344,13 +355,9 @@ sub configure
       my $service = $type . 'Service';
 
       if (exists $arg->{$host} && !exists $arg->{$service}) {
-         local $_ = $arg->{$host};
-         defined or next;
-         local ( $1, $2 ); # Placate a taint-related bug; [perl #67962]
-         if (/\A\[($IPv6_re)\](?::([^\s:]*))?\z/o || /\A([^\s:]*):([^\s:]*)\z/) {
-            $arg->{$host}    = $1;
-            $arg->{$service} = $2 if defined $2 && length $2;
-         }
+         defined $arg->{$host} or next;
+         ( $arg->{$host}, my $s ) = $self->split_addr( $arg->{$host} );
+         $arg->{$service} = $s if defined $s;
       }
    }
 
@@ -366,20 +373,22 @@ sub _configure
    my @localinfos;
    my @peerinfos;
 
-   $hints{flags} = $AI_ADDRCONFIG;
+   if( defined $arg->{GetAddrInfoFlags} ) {
+      $hints{flags} = $arg->{GetAddrInfoFlags};
+   }
+   else {
+      $hints{flags} = $AI_ADDRCONFIG;
+   }
 
-   # Check for definedness of args, but delete them anyway even if they're
-   # not defined. Then the only remaining keys will be unrecognised ones.
-
-   if( defined( my $family = delete $arg->{Family} ) ) {
+   if( defined( my $family = $arg->{Family} ) ) {
       $hints{family} = $family;
    }
 
-   if( defined( my $type = delete $arg->{Type} ) ) {
+   if( defined( my $type = $arg->{Type} ) ) {
       $hints{socktype} = $type;
    }
 
-   if( defined( my $proto = delete $arg->{Proto} ) ) {
+   if( defined( my $proto = $arg->{Proto} ) ) {
       unless( $proto =~ m/^\d+$/ ) {
          my $protonum = getprotobyname( $proto );
          defined $protonum or croak "Unrecognised protocol $proto";
@@ -403,7 +412,7 @@ sub _configure
       $hints{socktype} = SOCK_DGRAM  if $hints{protocol} == IPPROTO_UDP;
    }
 
-   if( my $info = delete $arg->{LocalAddrInfo} ) {
+   if( my $info = $arg->{LocalAddrInfo} ) {
       ref $info eq "ARRAY" or croak "Expected 'LocalAddrInfo' to be an ARRAY ref";
       @localinfos = @$info;
    }
@@ -430,17 +439,15 @@ sub _configure
          return;
       }
    }
-   delete $arg->{LocalHost};
-   delete $arg->{LocalService};
 
-   if( my $info = delete $arg->{PeerAddrInfo} ) {
+   if( my $info = $arg->{PeerAddrInfo} ) {
       ref $info eq "ARRAY" or croak "Expected 'PeerAddrInfo' to be an ARRAY ref";
       @peerinfos = @$info;
    }
    elsif( defined $arg->{PeerHost} or defined $arg->{PeerService} ) {
-      defined( my $host = delete $arg->{PeerHost} ) or
+      defined( my $host = $arg->{PeerHost} ) or
          croak "Expected 'PeerHost'";
-      defined( my $service = delete $arg->{PeerService} ) or
+      defined( my $service = $arg->{PeerService} ) or
          croak "Expected 'PeerService'";
 
       local $1; # Placate a taint-related bug; [perl #67962]
@@ -459,22 +466,20 @@ sub _configure
          return;
       }
    }
-   delete $arg->{PeerHost};
-   delete $arg->{PeerService};
 
    my @sockopts_enabled;
-   push @sockopts_enabled, SO_REUSEADDR if delete $arg->{ReuseAddr};
-   push @sockopts_enabled, SO_REUSEPORT if delete $arg->{ReusePort};
-   push @sockopts_enabled, SO_BROADCAST if delete $arg->{Broadcast};
+   push @sockopts_enabled, SO_REUSEADDR if $arg->{ReuseAddr};
+   push @sockopts_enabled, SO_REUSEPORT if $arg->{ReusePort};
+   push @sockopts_enabled, SO_BROADCAST if $arg->{Broadcast};
 
-   my $listenqueue = delete $arg->{Listen};
+   my $listenqueue = $arg->{Listen};
 
    croak "Cannot Listen with a PeerHost" if defined $listenqueue and @peerinfos;
 
-   my $blocking = delete $arg->{Blocking};
+   my $blocking = $arg->{Blocking};
    defined $blocking or $blocking = 1;
 
-   my $v6only = delete $arg->{V6Only};
+   my $v6only = $arg->{V6Only};
 
    # IO::Socket::INET defines this key. IO::Socket::IP always implements the
    # behaviour it requests, so we can ignore it, unless the caller is for some
@@ -482,9 +487,6 @@ sub _configure
    if( defined $arg->{MultiHomed} and !$arg->{MultiHomed} ) {
       croak "Cannot disable the MultiHomed parameter";
    }
-   delete $arg->{MultiHomed};
-
-   keys %$arg and croak "Unexpected keys - " . join( ", ", sort keys %$arg );
 
    my @infos;
    foreach my $local ( @localinfos ? @localinfos : {} ) {
@@ -510,6 +512,15 @@ sub _configure
       }
    }
 
+   if( !@infos and defined $hints{family} ) {
+      # If there was a Family hint then create a plain unbound, unconnected socket
+      @infos = ( {
+         family   => $hints{family},
+         socktype => $hints{socktype},
+         protocol => $hints{protocol},
+      } );
+   }
+
    # In the nonblocking case, caller will be calling ->setup multiple times.
    # Store configuration in the object for the ->setup method
    # Yes, these are messy. Sorry, I can't help that...
@@ -525,9 +536,9 @@ sub _configure
 
    ${*$self}{io_socket_ip_errors} = [ undef, undef, undef ];
 
-   if( $blocking ) {
-      $self->setup or return undef;
-   }
+   # ->setup is allowed to return false in nonblocking mode
+   $self->setup or !$blocking or return undef;
+
    return $self;
 }
 
@@ -565,17 +576,15 @@ sub setup
       }
 
       if( defined( my $addr = $info->{peeraddr} ) ) {
-         # It seems that IO::Socket hides EINPROGRESS errors, making them look
-         # like a success. This is annoying here.
-         # Instead of putting up with its frankly-irritating intentional
-         # breakage of useful APIs I'm just going to end-run around it and
-         # call CORE::connect() directly
-         if( CORE::connect( $self, $addr ) ) {
+         if( $self->connect( $addr ) ) {
             $! = 0;
             return 1;
          }
 
-         return 0 if $! == EINPROGRESS or HAVE_MSWIN32 && $! == Errno::EWOULDBLOCK();
+         if( $! == EINPROGRESS or HAVE_MSWIN32 && $! == Errno::EWOULDBLOCK() ) {
+            ${*$self}{io_socket_ip_connect_in_progress} = 1;
+            return 0;
+         }
 
          ${*$self}{io_socket_ip_errors}[0] = $!;
          next;
@@ -584,8 +593,6 @@ sub setup
       return 1;
    }
 
-   $self->close;
-
    # Pick the most appropriate error, stringified
    $! = ( grep defined, @{ ${*$self}{io_socket_ip_errors}} )[0];
    $@ = "$!";
@@ -595,16 +602,47 @@ sub setup
 sub connect
 {
    my $self = shift;
-   return $self->SUPER::connect( @_ ) if @_;
 
-   $! = 0, return 1 if $self->fileno and defined $self->peername;
+   # It seems that IO::Socket hides EINPROGRESS errors, making them look like
+   # a success. This is annoying here.
+   # Instead of putting up with its frankly-irritating intentional breakage of
+   # useful APIs I'm just going to end-run around it and call CORE::connect()
+   # directly
 
-   if( $self->fileno ) {
-      # A connect has just failed, get its error value
-      ${*$self}{io_socket_ip_errors}[0] = $self->getsockopt( SOL_SOCKET, SO_ERROR );
+   return CORE::connect( $self, $_[0] ) if @_;
+
+   return 1 if !${*$self}{io_socket_ip_connect_in_progress};
+
+   # See if a connect attempt has just failed with an error
+   if( my $errno = $self->getsockopt( SOL_SOCKET, SO_ERROR ) ) {
+      delete ${*$self}{io_socket_ip_connect_in_progress};
+      ${*$self}{io_socket_ip_errors}[0] = $! = $errno;
+      return $self->setup;
    }
 
-   return $self->setup;
+   # No error, so either connect is still in progress, or has completed
+   # successfully. We can tell by trying to connect() again; either it will
+   # succeed or we'll get EISCONN (connected successfully), or EALREADY
+   # (still in progress). This even works on MSWin32.
+   my $addr = ${*$self}{io_socket_ip_infos}[${*$self}{io_socket_ip_idx}]{peeraddr};
+
+   if( $self->connect( $addr ) or $! == EISCONN ) {
+      delete ${*$self}{io_socket_ip_connect_in_progress};
+      $! = 0;
+      return 1;
+   }
+   else {
+      $! = EINPROGRESS;
+      return 0;
+   }
+}
+
+sub connected
+{
+   my $self = shift;
+   return defined $self->fileno &&
+          !${*$self}{io_socket_ip_connect_in_progress} &&
+          defined getpeername( $self ); # ->peername caches, we need to detect disconnection
 }
 
 =head1 METHODS
@@ -627,6 +665,22 @@ sub _get_host_service
    return ( $host, $service );
 }
 
+sub _unpack_sockaddr
+{
+   my ( $addr ) = @_;
+   my $family = sockaddr_family $addr;
+
+   if( $family == AF_INET ) {
+      return ( Socket::unpack_sockaddr_in( $addr ) )[1];
+   }
+   elsif( defined $AF_INET6 and $family == $AF_INET6 ) {
+      return ( Socket::unpack_sockaddr_in6( $addr ) )[1];
+   }
+   else {
+      croak "Unrecognised address family $family";
+   }
+}
+
 =head2 ( $host, $service ) = $sock->sockhost_service( $numeric )
 
 Returns the hostname and service name of the local address (that is, the
@@ -652,7 +706,7 @@ sub sockhost_service
 
 =head2 $addr = $sock->sockhost
 
-Return the numeric form of the local address
+Return the numeric form of the local address as a textual representation
 
 =head2 $port = $sock->sockport
 
@@ -674,6 +728,14 @@ sub sockport { my $self = shift; ( $self->_get_host_service( $self->sockname, NI
 sub sockhostname { my $self = shift; ( $self->_get_host_service( $self->sockname, 0, NIx_NOSERV ) )[0] }
 sub sockservice  { my $self = shift; ( $self->_get_host_service( $self->sockname, 0, NIx_NOHOST ) )[1] }
 
+=head2 $addr = $sock->sockaddr
+
+Return the local address as a binary octet string
+
+=cut
+
+sub sockaddr { my $self = shift; _unpack_sockaddr $self->sockname }
+
 =head2 ( $host, $service ) = $sock->peerhost_service( $numeric )
 
 Returns the hostname and service name of the peer address (that is, the
@@ -697,7 +759,7 @@ sub peerhost_service
 
 =head2 $addr = $sock->peerhost
 
-Return the numeric form of the peer address
+Return the numeric form of the peer address as a textual representation
 
 =head2 $port = $sock->peerport
 
@@ -719,13 +781,21 @@ sub peerport { my $self = shift; ( $self->_get_host_service( $self->peername, NI
 sub peerhostname { my $self = shift; ( $self->_get_host_service( $self->peername, 0, NIx_NOSERV ) )[0] }
 sub peerservice  { my $self = shift; ( $self->_get_host_service( $self->peername, 0, NIx_NOHOST ) )[1] }
 
+=head2 $addr = $peer->peeraddr
+
+Return the peer address as a binary octet string
+
+=cut
+
+sub peeraddr { my $self = shift; _unpack_sockaddr $self->peername }
+
 # This unbelievably dodgy hack works around the bug that IO::Socket doesn't do
 # it
 #    https://rt.cpan.org/Ticket/Display.html?id=61577
 sub accept
 {
    my $self = shift;
-   my ( $new, $peer ) = $self->SUPER::accept or return;
+   my ( $new, $peer ) = $self->SUPER::accept( @_ ) or return;
 
    ${*$new}{$_} = ${*$self}{$_} for qw( io_socket_domain io_socket_type io_socket_proto );
 
@@ -746,6 +816,21 @@ sub socket
    dup2( $tmph->fileno, $self->fileno ) or die "Unable to dup2 $tmph onto $self - $!";
 }
 
+# Versions of IO::Socket before 1.35 may leave socktype undef if from, say, an
+#   ->fdopen call. In this case we'll apply a fix
+BEGIN {
+   if( $IO::Socket::VERSION < 1.35 ) {
+      *socktype = sub {
+         my $self = shift;
+         my $type = $self->SUPER::socktype;
+         if( !defined $type ) {
+            $type = $self->sockopt( Socket::SO_TYPE() );
+         }
+         return $type;
+      };
+   }
+}
+
 =head1 NON-BLOCKING
 
 If the constructor is passed a defined but false value for the C<Blocking>
@@ -813,6 +898,8 @@ C<PF_INET6>.
 For another example using C<IO::Poll> and C<Net::LibAsyncNS>, see the
 F<examples/nonblocking_libasyncns.pl> file in the module distribution.
 
+=cut
+
 =head1 C<PeerHost> AND C<LocalHost> PARSING
 
 To support the C<IO::Socket::INET> API, the host and port information may be
@@ -841,6 +928,67 @@ service name and number, in a form such as
 In this case, the name (C<http>) will be tried first, but if the resolver does
 not understand it then the port number (C<80>) will be used instead.
 
+=head2 ( $host, $port ) = IO::Socket::IP->split_addr( $addr )
+
+Utility method that provides the parsing functionality described above.
+Returns a 2-element list, containing either the split hostname and port
+description if it could be parsed, or the given address and C<undef> if it was
+not recognised.
+
+ IO::Socket::IP->split_addr( "hostname:http" )
+                              # ( "hostname",  "http" )
+
+ IO::Socket::IP->split_addr( "192.0.2.1:80" )
+                              # ( "192.0.2.1", "80"   )
+
+ IO::Socket::IP->split_addr( "[2001:db8::1]:80" )
+                              # ( "2001:db8::1", "80" )
+
+ IO::Socket::IP->split_addr( "something.else" )
+                              # ( "something.else", undef )
+
+=cut
+
+sub split_addr
+{
+   shift;
+   my ( $addr ) = @_;
+
+   local ( $1, $2 ); # Placate a taint-related bug; [perl #67962]
+   if( $addr =~ m/\A\[($IPv6_re)\](?::([^\s:]*))?\z/ or
+       $addr =~ m/\A([^\s:]*):([^\s:]*)\z/ ) {
+      return ( $1, $2 ) if defined $2 and length $2;
+      return ( $1, undef );
+   }
+
+   return ( $addr, undef );
+}
+
+=head2 $addr = IO::Socket::IP->join_addr( $host, $port )
+
+Utility method that performs the reverse of C<split_addr>, returning a string
+formed by joining the specified host address and port number. The host address
+will be wrapped in C<[]> brackets if required (because it is a raw IPv6
+numeric address).
+
+This can be especially useful when combined with the C<sockhost_service> or
+C<peerhost_service> methods.
+
+ say "Connected to ", IO::Socket::IP->join_addr( $sock->peerhost_service );
+
+=cut
+
+sub join_addr
+{
+   shift;
+   my ( $host, $port ) = @_;
+
+   $host = "[$host]" if $host =~ m/:/;
+
+   return join ":", $host, $port if defined $port;
+   return $host;
+}
+
 =head1 C<IO::Socket::INET> INCOMPATIBILITES
 
 =over 4
